local revision = tonumber(string.sub("$Revision: 55 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end

-- matches an AA with the spell that triggered it, based on closest match to 30%
function rss.MatchAA(amount)
	local mindiff = amount;                                         -- Match this AA to the spell that triggered it, based on the 30% relationship
	local sMin = nil;
	local vMin = nil;
	for s, v in pairs(rss.playerState.lastCast) do
		if math.abs(v.amount*0.3 - amount) < mindiff then
			mindiff = math.abs(v.amount*0.3 - amount);
			vMin = v;
			sMin = s;
		end
	end
	return sMin, vMin;
end

function rss.CalcCopy(record, amount, incrementalIntGain, masteryGain, incrementalMasteryGain, critical)
	if critical then
		record.rawHealing[1] = record.rawHealing[1] + amount;
	else
		record.rawHealing[2] = record.rawHealing[2] + amount;
	end
	record.incrementalInt = record.incrementalInt + incrementalIntGain;
	record.masteryTotal = record.masteryTotal + masteryGain;
	record.incrementalMastery = record.incrementalMastery + incrementalMasteryGain;
end

function rss.Leak(reason, healed)
	print("RSS: Leakage detected");
	if not rss.fight.leakage[reason] then
		rss.fight.leakage[reason] = 0.0;
	end
	rss.fight.leakage[reason] = rss.fight.leakage[reason] + healed;
end

local spellSchools = { }; -- lookup table to cache expensive logarithm calculation 

function rss.CombatLogEntry(...)
	local playerState = rss.playerState;
	local lastCast = playerState.lastCast;
	local fight = rss.fight;
	local timestamp, combatEvent, hideCaster, 
	      sourceGUID, sourceName, sourceFlags, sourceRaidFlags, 
	      destGUID, destName, destFlags, destRaidFlags = select(1, ...);
	
	-- watch for the comings and goings of Spirit Link Totem
	if sourceGUID == playerState.GUID and combatEvent and strfind(combatEvent, 'SPELL_SUMMON') then
		local spellId, spellName, spellSchool = select(12, ...);
		if spellId == rss.SUMMON_SPIRIT_LINK then 
			playerState.spiritLinkGUID = destGUID;
		end
		return;
	end
	
	if combatEvent and strfind(combatEvent, 'UNIT_DESTROYED') and destGUID == playerState.spiritLinkGUID then
		playerState.spiritLinkGUID = nil;
		return;
    end

    -- handle power gains like resurgence
    if sourceGUID == playerState.GUID and combatEvent and combatEvent == "SPELL_ENERGIZE" then
        local spellId, spellName, spellSchool = select(12, ...)
        local amount, powerType = select(15, ...)
        if spellId == rss.RESURGENCE then
            fight.resurgence = fight.resurgence + amount
        end
        return;
    end
	
	-- deal with the Flame Shock glyph
	if sourceGUID == playerState.GUID and combatEvent and strfind(combatEvent, '_DAMAGE') then
		local spellId = select(12,...);
		if spellId == rss.FLAME_SHOCK then
			playerState.lastFlameShock.combatEvent = combatEvent;
			playerState.lastFlameShock.critical = select(21,...);
		end
		return;
	end

	-- The remainder only looks at 
	-- 1) Heals
	-- 2) Cast by us or our SLT
	-- 3) On a target whose health we know
	local healthTable = rss.healthTable;
	local spiritLinkHeal = playerState.spiritLinkGUID and sourceGUID == playerState.spiritLinkGUID;
	if (sourceGUID == playerState.GUID or spiritLinkHeal) and combatEvent and strfind(combatEvent, '_HEAL') and healthTable[destGUID] then
		-- Check for boss encounter by looking at the target of anybody we heal
		if ( (not fight.isBoss) or (fight.encounterName == "Unknown") ) then
			rss.CheckBoss(destName);
		end

		-- Check for first action of combat
		if fight.startTime <= 0 then
            fight.startTime = timestamp;
		end
        fight.totalTime = timestamp - fight.startTime;
		
		-- We healed somebody
		local spellId, spellName, spellSchool = select(12, ...)
		local amount, overhealing, absorbed, critical = select(15, ...)
		if spellId == rss.FLAMES_OF_LIFE and playerState.lastFlameShock then
			critical = playerState.lastFlameShock.critical;
		elseif spellId == rss.GLYPH_OF_HEALING_WAVE and lastCast[rss.HEALING_WAVE] then
			critical = lastCast[rss.HEALING_WAVE].critical;
		end
		
		local healed = math.max(0.,amount - overhealing);                       -- ignore overhealing
        fight.effectiveTotal = fight.effectiveTotal + healed;                   -- accumulate total effective throughput
		
		if spiritLinkHeal then 													-- no stat effects on Spirit Link totem heals
			return;
        end

        -- count the number of times we healed with non-periodic spells
        if combatEvent == "SPELL_HEAL" then
            local spellId_count = fight.spells[spellId] or 0
            spellId_count = spellId_count + 1
            fight.spells[spellId] = spellId_count
        end

		-- match AA once for everything that follows
		local aaSpell = nil;
		local aaCast  = nil;
		if spellId == rss.ANCESTRAL_AWAKENING then
			aaSpell, aaCast = rss.MatchAA(amount);
			if aaCast == nil then
				rss.Leak("AA match not found", healed);
				return;
			end
		end
		
		local deficit = healthTable[destGUID].maxHealth - healthTable[destGUID].currentHealth;		
		-- deal with spell copies from Ascendance and Ancestral Guidance
		if spellId == rss.RESTORATIVE_MISTS or spellId == rss.ANCESTRAL_GUIDANCE then
			if playerState.ascendance and spellId == rss.RESTORATIVE_MISTS then
				local ascendance = playerState.ascendance;
				if (overhealing == 0) then
					ascendance.effectiveCopies = ascendance.effectiveCopies + healed;
				end
				ascendance.incrementalCritical[1] = ascendance.incrementalCritical[1] 
					+ (0.01/rss.RATING_TO_CRIT)*(math.min(amount, deficit) - math.min(amount/playerState.criticalMultiplier, deficit));
				ascendance.incrementalCritical[2] = ascendance.incrementalCritical[2]
				    + (0.01/rss.RATING_TO_CRIT)*(math.min(amount*playerState.criticalMultiplier, deficit) - math.min(amount, deficit));
			elseif playerState.ancestralGuidance and spellId == rss.ANCESTRAL_GUIDANCE then
				local guidance = playerState.ancestralGuidance;
				if (overhealing == 0) then
					guidance.effectiveCopies = guidance.effectiveCopies + healed;	
				end
				guidance.incrementalCritical[1] = guidance.incrementalCritical[1] 
					+ (0.01/rss.RATING_TO_CRIT)*(math.min(amount, deficit) - math.min(amount/playerState.criticalMultiplier, deficit));
				guidance.incrementalCritical[2] = guidance.incrementalCritical[2]
				    + (0.01/rss.RATING_TO_CRIT)*(math.min(amount*playerState.criticalMultiplier, deficit) - math.min(amount, deficit));				
			else
				rss.Leak("Missing originals for copy", healed);  -- derp, we have no record of the spell(s) that this heal was copied from
			end
			return; -- nothing more to do for now; stats affect the spell that was copied, not the copy itself
		end
		
		-- Recognize consumables so they can be (mostly) ignored
		local consumable = tContains(rss.CONSUMABLES, spellId);
		local bandage;
		if consumable then
			bandage = tContains(rss.BANDAGES, spellId);
		end
		-- Recognize totems
		local totem;
		if rss.SPELLS[spellId] and rss.SPELLS[spellId].totem then
			totem = true;
		end
		
		-- Deal with Intellect
		local mySpellPower = playerState.healingBonus;
		local incrementalInt = 0.0;		
		if rss.SPELLS[spellId] and rss.SPELLS[spellId][combatEvent] then
			local c = rss.SPELLS[spellId][combatEvent];
			local baseAmt = (c[1]+c[2])/2 + mySpellPower*c[3];
			incrementalInt = (1.1*c[3]/baseAmt) * amount;   -- the 1.1 reflects our 10% spellpower buff
		elseif spellId == rss.ANCESTRAL_AWAKENING then
			incrementalInt = 0.3*aaCast.incrementalIntellect;
		elseif spellId == rss.GLYPH_OF_HEALING_WAVE then
			if lastCast[rss.HEALING_WAVE] then
				incrementalInt = 0.2*lastCast[rss.HEALING_WAVE].incrementalIntellect;
			else
				rss.Leak("Missing original HW for glyph heal", healed);
				return;
			end
		elseif spellId == rss.FLAMES_OF_LIFE then
			mySpellPower = playerState.damageBonus[rss.SCHOOL_FIRE];
			local c = SPELLS[spellId][playerState.lastFlameShock.combatEvent];
			local baseAmt = (c[1]+c[2])/2 + mySpellPower*c[3];
			incrementalInt = (1.1*c[3]/baseAmt) * amount;
		elseif spellId ~= rss.GIFT_OF_THE_NAARU and spellId ~= rss.CLEANSING_WATERS and not consumable then
			if not rss.unknownSpells[spellId] then
				print("Unrecognized healing entry: "..spellName.." "..spellId.." "..combatEvent);
				rss.unknownSpells[spellId] = { name = spellName , count = 0 };
			end
			rss.unknownSpells[spellId].count = rss.unknownSpells[spellId].count + 1;
			consumable = true;  -- assume no stat benefits other than crit
		end
		if overhealing == 0 then
			fight.incrementalIntellect = fight.incrementalIntellect + incrementalInt;
		end
		
		-- Deal with Mastery
		local masteryGain = 0.0;
		local incrementalMasteryGain = 0.0;
		local effectiveMasteryGain = 0.0;
		local hpratio = 0.0;
		local unitHealthTable = healthTable[destGUID];							-- find last known values of health and max health
		if (not consumable) and rss.SPELLS[spellId] and (not rss.SPELLS[spellId].noMastery) then
			if spellId == rss.ANCESTRAL_AWAKENING then
				masteryGain = 0.3 * aaCast.masteryGain;
				effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);
				incrementalMasteryGain = 0.3 * aaCast.incrementalMastery;
				hpratio = aaCast.hpRatio;
			elseif spellId == rss.GLYPH_OF_HEALING_WAVE then
				masteryGain = 0.2 * lastCast[rss.HEALING_WAVE].masteryGain;
				effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);
				incrementalMasteryGain = 0.2 * lastCast[rss.HEALING_WAVE].incrementalMastery;
				hpratio = lastCast[rss.HEALING_WAVE].hpRatio;			
			else
				local myMasteryBonus = playerState.masteryBonus;
				hpratio = 1 - (unitHealthTable.currentHealth/unitHealthTable.maxHealth);
				local healBaseAmount = amount/(1+hpratio*myMasteryBonus);               -- the size of the heal with 0% Mastery
				masteryGain = amount - healBaseAmount;                            		-- increase from base amount from Mastery
				effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);  		-- overhealing is not useful
				incrementalMasteryGain = (0.01/rss.RATING_TO_MASTERY) * hpratio * healBaseAmount;     -- how much would +1 Mastery Rating increase this heal?
			end
			fight.weightedTotal = fight.weightedTotal + healed * hpratio
			fight.effectiveSubTotal = fight.effectiveSubTotal + healed
			fight.masteryTotal = fight.masteryTotal + effectiveMasteryGain;						-- accumulate gain from current Mastery
			if (overhealing == 0) then
				fight.incrementalMastery = fight.incrementalMastery + incrementalMasteryGain;       -- accumulate gain from +1 Mastery
			end	
		end
		
		-- Deal with Crit
		local schoolIndex;
		if rss.SPELLS[spellId] and rss.SPELLS[spellId].school then
			schoolindex = rss.SPELLS[spellId].school;
		else
			if not spellSchools[spellSchool] then
				spellSchools[spellSchool] =  math.floor(1+log(spellSchool)/log(2));
			end
			schoolIndex = spellSchools[spellSchool];
		end
		
		local myCriticalChance = playerState.criticalChance[schoolIndex] or playerState.criticalChance[rss.SCHOOL_NATURE];
		local thisCast = nil;                                                   -- record this as the most recent cast of this spell, for matching with AA
		if spellId ~= rss.ANCESTRAL_AWAKENING then                                  -- unless it is AA healing, which can't trigger AA
			thisCast = { };
			thisCast.amount = amount;
			if spellId == rss.FLAMES_OF_LIFE then
				thisCast.periodic = strfind(playerState.lastFlameShock.combatEvent, 'PERIODIC');
			else
				thisCast.periodic = strfind(combatEvent, 'PERIODIC_HEAL');
			end
			thisCast.hpRatio = hpratio;
			thisCast.critChance = myCriticalChance;
			thisCast.spellPower = mySpellPower;
			thisCast.masteryGain = masteryGain;
			thisCast.incrementalMastery = incrementalMasteryGain;
			thisCast.incrementalIntellect = incrementalInt;
			thisCast.critical = false;
		end
		local critBonusAmount = 0.0;
		if critical or (spellId == rss.ANCESTRAL_AWAKENING) then                  
			if spellId == rss.HEALING_SURGE then                                -- is this a Healing Surge critical?
				if playerState.tidalWaves then              					-- tested: even if this cast consumed the last TW stack, we still have the buff at this moment
					myCriticalChance = math.min(1.0, myCriticalChance + 0.3);   -- adjust HS critical chance for TW
					thisCast.critChance = myCriticalChance;                     -- record the buffed value of critical chance for use with any AA triggered by this heal
				end
			end
			if critical then   -- this cast crit, so figure out how much we gained over the base heal
				thisCast.critical = true;                                       -- record the cast as critical so we can add the triggered AA to the critical category
				local critBaseAmount = amount/playerState.criticalMultiplier;               
				critBonusAmount = amount - critBaseAmount;
                fight.criticalTotal = fight.criticalTotal + math.max(0.0, critBonusAmount - overhealing);  -- record the effective gain from this crit
                fight.incrementalCritical[1] = fight.incrementalCritical[1] + (0.01/rss.RATING_TO_CRIT)*math.max(0.0, critBonusAmount - overhealing);
			else -- Ancestral Awakening                                      
				if aaCast.critical then                          			   -- if the trigger for this AA was a crit (rather than the set bonus)..
					local nonCritAA = 0;                                       -- size of a possible non-crit AA from original heal; zero unless set bonus present
					if playerState.aaSetBonusT15 then
						nonCritAA = 0.5 * rss.SPELLS[spellId].copy * aaCast.amount / playerState.criticalMultiplier     -- with set bonus, half of 30% of the non-crit heal amount expected
					end
					fight.criticalTotal = fight.criticalTotal + math.max(0, healed - nonCritAA); -- count it as gain from critical
					fight.incrementalCritical[1] = fight.incrementalCritical[1] + (0.01/rss.RATING_TO_CRIT)*math.max(0, healed - nonCritAA);
				else  -- this AA was triggered by the set bonus; remove the current (non-crit) AA amount from the gain we added previously
					fight.incrementalCritical[2] = math.max(0, fight.incrementalCritical[2] - (0.01/rss.RATING_TO_CRIT)*healed);
				end
				return; -- nothing more to do with AA's
			end
		elseif not bandage then -- this is not a critical or an AA, but it could have crit
			local criticalAmount = amount * playerState.criticalMultiplier;                -- how big would the raw heal be if it crit?
			critBonusAmount = criticalAmount - amount;
			if overhealing == 0 then										   -- no benefit to the target from critting on an overheal
				local criticalHeal = math.min(criticalAmount, deficit);
                fight.incrementalCritical[2] = fight.incrementalCritical[2] + (0.01/rss.RATING_TO_CRIT)*(criticalHeal - healed);
			end
			-- add healing from the hypothetical AA, if applicable; we assume it would hit the friendly on our list with the largest health deficit ("most injured")
			if rss.SPELLS[spellId] and rss.SPELLS[spellId].aaTrigger and not thisCast.periodic then
				local criticalAA = rss.SPELLS[rss.ANCESTRAL_AWAKENING].copy * criticalAmount;    -- how big would the AA heal be if it crit?
				local maxDeficit = 0;
				for guid, t in pairs(healthTable) do
					local aaDeficit = math.max(0, t.maxHealth - t.currentHealth);
					if guid == destGUID then                                  -- account for the amount we *would* have healed the target...
						aaDeficit = math.max(0, aaDeficit - criticalAmount);
					end
					maxDeficit = math.max(maxDeficit, aaDeficit);
				end
				criticalAA = math.min(criticalAA, maxDeficit);
                fight.incrementalCritical[2] = fight.incrementalCritical[2] + (0.01/rss.RATING_TO_CRIT)*criticalAA;
			end
		end
		if thisCast ~= nil and rss.SPELLS[spellId] and rss.SPELLS[spellId].aaTrigger and not thisCast.periodic then         -- store information on this cast for AA matching and accounting
			lastCast[spellId] = thisCast;
		end
	-- done with Mastery and Crit, now store info for active copy buffs
		if (playerState.ascendanceAura or playerState.guidanceAura) and not totem then
			if playerState.ascendance then
				rss.CalcCopy(playerState.ascendance, amount, incrementalInt, masteryGain, incrementalMasteryGain, critical);
			end
			if playerState.ancestralGuidance and (not thisCast.periodic) and (consumable or not SPELLS[spellId] or not SPELLS[spellId].noAG) then
				rss.CalcCopy(playerState.ancestralGuidance, amount, incrementalInt, masteryGain, incrementalMasteryGain, critical)
			end
		end
	end
end
